#include "datatype.h"
#include "confprot.h"
#include "processStimuli.h"


int	processStimuli(ConfProtData CPD)
{
	int	result;
	char	pdu[MAXPDUSIZE];

	MarkCoverage();
	result = read(CPD->stdinhandle, (char *) & pdu, MAXPDUSIZE);

#ifndef M06
	if ((result < 0)) {
		MarkCoverage();
#endif
		WriteMSG("Error in read operation!\n");
		return(-20);
#ifndef M06
	}
	MarkCoverage();
#endif

#ifndef M07
	if ((result == 0)) {
		MarkCoverage();
#endif
		WriteMSG("EOF detected!\n");
		return(-21);
#ifndef M07
	}
	MarkCoverage();
#endif

#ifdef M08
		pdu[0] = STOP; /* Perform STOP pdu */
#endif

#ifdef M54
		pdu[0] = 255; /* Force unknow pdu packet */
#endif

#ifdef M55
		pdu[0] = ANSWER; /* Force answer pdu packet */
#endif

	/*
	  * switch over pdu type field
	  */
	switch (pdu[0]) {
#ifdef M08
	case STOP:
		{
		MarkCoverage();
			WriteMSG("STOP pdu detected!\n");
			return(-22);
		}
#endif

	case JOIN:
		 {
			int	p = 0;

			MarkCoverage();
#ifndef M10
			if ((CPD->active)) {
				MarkCoverage();
#endif
				break;
#ifndef M10
			}
			MarkCoverage();
#endif


#ifndef M60
			/*
			* Get nickname...
			*/
			memcpy(CPD->nickname, &pdu[1], MAXNICKNAME);

			/*
			* Get conference...
			*/
			memcpy(CPD->conference, &pdu[MAXNICKNAME+1], MAXCONFERENCE);
#endif

			/*
			* Activate conference...
			*/
#ifndef M26
			CPD->active = 1;
#endif

			/*
			* Broadcast to potential partners
			*/
			for (p = 2 ; p < CPD->nusers ; p++) {
				MarkCoverage();
#ifdef M21
				break;
#endif
				transmit(CPD, pdu, JOINPDUSIZE, JOIN, p);
#ifdef M39
				transmit(CPD, pdu, JOINPDUSIZE, JOIN, p);
#endif
			}
			MarkCoverage();

#ifdef M38
			pdu[0] = LEAVE;

			for (p = 2 ; p < CPD->nusers ; p++) {
				MarkCoverage();
				transmit(CPD, pdu, LEAVEPDUSIZE, LEAVE, p);
			}
			MarkCoverage();
			CPD->active = 0;
#endif
			break;
		}

	case DATA:
		 {
			int	p = 0;

			MarkCoverage();
#ifndef M11
			if ((!CPD->active)) {
				MarkCoverage();
#endif
				break;
#ifndef M11
			}
			MarkCoverage();
#endif

			/*
			      * Transmit msg to all partners
			      */
			for (p = 0 ; p < CPD->npartners ; p++) {
				int	pdusize = 1 + 1 + pdu[1];
				MarkCoverage();

#ifdef M22
				break;
#endif

#ifdef M44
				pdusize = MAXPDUSIZE;
#endif
				transmit(CPD, pdu, pdusize, DATA, CPD->p2cid[p]);
#ifdef M40
				transmit(CPD, pdu, pdusize, DATA, CPD->p2cid[p]);
#endif
			}
			MarkCoverage();

			break;
		}

	case ANSWER:
		 {
			MarkCoverage();
			WriteMSG("ANSWER pdu detected!\n");
			break;
		}

	case LEAVE:
		 {
			int	p = 0;

			MarkCoverage();
#ifndef M12
			if ((!CPD->active)) {
				MarkCoverage();
#endif
				break;
#ifndef M12
			}
			MarkCoverage();
#endif
			/*
			* Clear active conference...
			*/
#ifndef M24
			CPD->active = 0;
#endif

			memcpy(&pdu[1], CPD->nickname, MAXNICKNAME);
			memcpy(&pdu[1+MAXNICKNAME], CPD->conference, MAXCONFERENCE);

			/*
			      * Transmit to all partners
			      */
			for (p = 0 ; p < CPD->npartners ; p++) {
				MarkCoverage();
#ifdef M23
				break;
#endif
				transmit(CPD, pdu, LEAVEPDUSIZE, LEAVE, CPD->p2cid[p]);
#ifdef M41
				transmit(CPD, pdu, LEAVEPDUSIZE, LEAVE, CPD->p2cid[p]);
#endif
			}
			MarkCoverage();

			/*
			* Clear partnerset...
			*/
#ifndef M25
			CPD->npartners = 0;
#endif
			break;
		}
	default:
		{
			MarkCoverage();
			WriteMSG("Unknown stimuli detected!\n");
			break;
		}
	}

	MarkCoverage();
	return(0);
}


