#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define maxcovtable 5*1024

struct covtabletype {
  int freq;
  char filename[80];
};

struct covtabletype covtable[maxcovtable];

int covtotalblocks=0;
int covmarkedblocks=0;
char coveragedatafile[80];
char coverage[80];
int CUMMULATIVE_COVERAGE=0;
int INCREMENTAL_COVERAGE=0;
int NO_COVERAGE=0;

static int print_coverage(int i);

int init_coverage(int argc, char *argv[])
{
  int i=0;
  int n=0;
  char *chp;
  FILE *f;
  int mode=0;
  int j=0;

  /* clear coverage data */
  for (i=0; i<maxcovtable; i++) {
    covtable[i].filename[0]='\0';
    covtable[i].freq=0;
  }
  
  chp = getenv("COVERAGE");
  if (chp) {
    strcpy(coverage, chp);
  } else {
    strcpy(coverage, "");
  }
  
  /* get coverage data filename from environment */
  chp = getenv("COVDATAFILE");
  if (chp) {
    strcpy(coveragedatafile, chp);
  } else {
    strcpy(coveragedatafile, "");
  }
  
  mode = 0;
  for (j = 1; j < argc; j++) {
    switch (mode) {
      case 0 : {
        if (strcmp(argv[j], "-d") == 0) {
          mode = 1;
        } else if (strcmp(argv[j], "-s") == 0) {
          mode = 2;
        }
        break;
      }
      case 1 : {
        strcpy(coveragedatafile, argv[j]);
        mode = 0;
        break;
      }
      case 2 : {
        strcpy(coverage, argv[j]);
        mode = 0;
        break;
      }
      default : {
        mode = 0;
        break;
      }
    }
  }
  
  if (strchr(coverage, 'C')) {
    CUMMULATIVE_COVERAGE=1;
  } else {
    CUMMULATIVE_COVERAGE=0;
  }
  if (strchr(coverage, 'I')) {
    INCREMENTAL_COVERAGE=1;
  } else {
    INCREMENTAL_COVERAGE=0;
  }
  if (strchr(coverage, 'D')) {
    NO_COVERAGE=1;
  } else {
    NO_COVERAGE=0;
  }
  
  if (NO_COVERAGE==1) {return(0);}
  
  f=fopen(coveragedatafile, "r");
  
  if (f==NULL) {
    fprintf(stderr,"COVERAGEDATAFILE not loaded! (%s)\n", coveragedatafile);
    exit(-1);
  }
  
  covtotalblocks=0; 
  while ((n=fscanf(f, "%s %d\n", 
  	covtable[covtotalblocks].filename, 
	&covtable[covtotalblocks].freq)) >= 1) {

  	if (covtable[covtotalblocks].freq>0) {
      covmarkedblocks++;
  	}
  	covtotalblocks++; 
  }
  fclose(f);
  print_coverage(-1);
  return(0);
}

int done_coverage()
{
  
  if (NO_COVERAGE==1) {return(0);}
  
  if (CUMMULATIVE_COVERAGE==1) {
    int i=0;
    FILE *f;
    f=fopen(coveragedatafile, "w");
    for (i=0;i<covtotalblocks;i++) {
      fprintf(f, "%s %d\n", covtable[i].filename, covtable[i].freq);
    }
    fclose(f);
  }
  return(0);
}

int MarkCov(char *f)
{
  int i=0;
  int b=0;
  int e=0;
  int x=0;
  
  
  if (NO_COVERAGE==1) {return(0);}
  
/*  for (i=0;i<covtotalblocks;i++) {
    if (strcmp(covtable[i].filename, f)==0) {

      if (covtable[i].freq==0) {
        covmarkedblocks++;
	    covtable[i].freq++;
      } else if (INCREMENTAL_COVERAGE==1) {
	    covtable[i].freq++;
  	    return(0);
      }
	  break; 
    }
  }
*/

	b = 0;
	e = covtotalblocks-1;
	i = (b + e) / 2;
	while ( b < e ) {
      if ((x=strcmp(covtable[i].filename, f)) > 0 ) {
        e = i-1;
      } else if (x < 0 ) {
        b = i+1;
      } else {
        e = i;
        b = i;
      }
	  i = (b + e) / 2;
	}
    if (b==e) {
      if (covtable[i].freq==0) {
        covmarkedblocks++;
	    covtable[i].freq++;
      } else if (INCREMENTAL_COVERAGE==1) {
	    covtable[i].freq++;
  	    return(0);
      }
    }
  
  if (i==covtotalblocks) {return(0);}
  
  print_coverage(i);
  return(0);
}

int print_coverage(int i) {
  fprintf(stdout, "LOG COVERAGE total=%d todo=%d done=%d percent=%f", 
  	covtotalblocks, 
  	covtotalblocks-covmarkedblocks, 
  	covmarkedblocks, 
  	(float)(covmarkedblocks*100)/covtotalblocks
  ); 
  if (i>=0) {
    fprintf(stdout, " id=%s", covtable[i].filename); 
  }
  fprintf(stdout, "\n"); 
  fflush(stdout);
  return(0);
}

int main(int argc, char *argv[]) {
  char filename[80];
  int freq;
  
  init_coverage(argc, argv);

  
  while (scanf("%s %d\n", filename, &freq)>=1) {
    MarkCov(filename);
  }
  
  done_coverage();
  
  return(0);
}
